"""Save output docstring to txt file."""

import argparse
import os
from typing import List

from docstripy.file_parser import parse_ranges


def extract_docstrs(dir_path: str) -> List[str]:
    """Extract all docstrings from files in dir_path."""
    all_docstrs = []
    for dir_path, _, file_names in os.walk(dir_path):
        for file_name in file_names:
            if file_name.endswith(".py"):
                path = os.path.join(dir_path, file_name)
                all_docstrs.append(f">>> PATH: {path}\n")
                with open(path, "r", encoding="utf-8") as file:
                    lines = file.readlines()
                ranges_docstr, _ = parse_ranges(lines)
                for range_docstr in ranges_docstr:
                    line_docstr = lines[range_docstr[0] : range_docstr[1]]
                    all_docstrs.extend(line_docstr)
    return all_docstrs


def get_args() -> dict:
    """Get command line args."""
    parser = argparse.ArgumentParser(description="Save output docstring to txt file.")
    parser.add_argument("input", type=str, help="Input dir path")
    parser.add_argument("output", type=str, help="Ouptout .txt file path")
    args = parser.parse_args()
    return {"input": args.input, "output": args.output}


def main() -> None:
    """Save output docstring to txt file."""
    args = get_args()
    all_docstrs = extract_docstrs(args["input"])
    os.makedirs(os.path.dirname(args["output"]), exist_ok=True)
    with open(args["output"], "w", encoding="utf-8") as file:
        file.writelines(all_docstrs)


if __name__ == "__main__":
    main()



# type: ignore
"""A test module.

Contains some classes for testing the parser.
"""


class MyClass:
    """My class.

    Parameters
    ----------
    attr:
        The attribute
    """

    # A comment
    cls_attr: 0  # some class attribute

    def __init__(self, attr: int = 0) -> None:
        self.attr = attr


class MyOtherClass:
    """My other class.

    Parameters
    ----------
    attr: int
        The attribute.
    """

    def forward(self, str_in: str) -> None:  # noqa
        return str_in



# type: ignore
"""A test module.

Contains some functions for testing the parser.
"""


def factorial(n):
    """Return factorial of n.

    Parameters
    ----------
    n : int
        Number to compute factorial of.

    Returns
    -------
    int
        Factorial of n.

    Raises
    ------
    ValueError
    """
    if n < 0:
        raise ValueError("Cannot compute factorial of negative number.")
    if n == 0:
        return 1
    return n * factorial(n - 1)


def fibonacci(n):
    """Return nth Fibonacci number."""
    if n < 0:
        raise ValueError("Cannot compute Fibonacci number of negative index.")
    if n == 0:
        return 0
    if n == 1:
        return 1
    print("""
    This is a test.
    """)
    message = """
    This is a second test.
    """
    print(message)
    message = """
    This is a third one.
    """
    print(message)
    return fibonacci(n - 1) + fibonacci(n - 2)


def square(n):
    r"""Return square.

    Use \escaped characters.
    """
    return n * n


def empty_function() -> None:  # noqa: D103
    pass



"""Test line break."""

import pytest_check as check

from docstripy.line_break import line_break


def test_line_break() -> None:
    """Test line break."""
    lines = [
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit.\n",
        "Sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.\n",
        "Ut enim ad minim veniam, quis nostrud. Exercitation ullamco laboris nisi\n",
        "ut aliquip ex ea commodo consequat...\n",
        "Excepteur sint occaecat cupidatat:\n",
        "- aute irure dolor\n",
        "- dolor in reprehenderit\n",
        "\n",
        "Duis aute irure dolor in reprehenderit in voluptate velit esse cillum!\n",
        "\n",
        "Dolore eu fugiat nulla pariatur.\n",
        "Excepteur sint occaecat cupidatat non proident, sunt in culpa qui\n",
        "officia deserunt mollit anim id est laborum.\n",
        "Et dolorum fuga...\n",
    ]
    expected_lines = [
        "Lorem ipsum dolor sit\n",
        "amet, consectetur adipiscing elit. Sed do eiusmod\n",
        "tempor incididunt ut labore et dolore magna\n",
        "aliqua. Ut enim ad minim veniam, quis nostrud.\n",
        "Exercitation ullamco laboris nisi ut aliquip ex ea\n",
        "commodo consequat... Excepteur sint occaecat\n",
        "cupidatat:\n",
        "- aute irure dolor\n",
        "- dolor in reprehenderit\n",
        "\n",
        "Duis aute irure dolor in reprehenderit in\n",
        "voluptate velit esse cillum!\n",
        "\n",
        "Dolore eu fugiat nulla pariatur. Excepteur sint\n",
        "occaecat cupidatat non proident, sunt in culpa qui\n",
        "officia deserunt mollit anim id est laborum.\n",
        "Et dolorum fuga...\n",
    ]
    lines = line_break(lines=lines, max_line_length=50, num_add_char=25)
    check.equal(len(lines), len(expected_lines))
    for i, line in enumerate(lines):
        check.equal(
            bytes(line, encoding="utf-8"),
            bytes(expected_lines[i], encoding="utf-8"),
            f"Error with line {i}",
        )



# type: ignore
"""A test module.

Contains some functions for testing the parser.
"""


class TestAttr:
    """Test class for attributes.

    This is a function where the ending quotes
    are in the same line as the title.

    :cvar integer: The number
    :ivar str: The string
    :var bool: The boolean
    """

    integer = 0

    def __init__(self) -> None:
        self.str = "string"
        self.bool = True



"""Configuration file for the Sphinx documentation builder."""
# pylint: disable=all
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import sys

from setuptools_scm import get_version

sys.path.insert(0, os.path.abspath('../'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'DocstriPy'
copyright = '2024, Valentin Goldite'  # noqa A001
author = 'Valentin Goldite'
try:
    release = get_version()
except:  # noqa E722
    release = get_version(root='..', relative_to=__file__)
release = ".".join(release.split(".")[:3]) + ".dev" if "dev" in release else ""

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
]

master_doc = "index"
autoapi_type = 'python'
autoapi_dirs = ["docstripy"]

autodoc_default_options = {
    "member-order": "bysource",
    "undoc-members": True,
}

add_module_names = False
autoclass_content = "both"
napoleon_use_param = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
}

templates_path = ['_templates']
exclude_patterns = ['_build']

rst_prolog = """
.. |default| raw:: html

    <div class="default-value-section">""" + \
    ' <span class="default-value-label">Default:</span>'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']
html_css_files = ["custom.css"]

html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "valentingol",  # Username
    "github_repo": "docstripy",  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}



"""Signature related functions."""

from typing import List, Optional

from docstripy.lines_routines import clean_leading_empty, find_indent
from docstripy.parse_doc.parse_signature import parse_signature


def merge_docstr_signature(
    sections_docstr: dict,
    lines_signature: List[str],
) -> dict:
    """Merge docstring and definition ranges."""
    if not lines_signature:
        return sections_docstr
    fn_name, rtypes, args = parse_signature(lines_signature)

    if "_title" not in sections_docstr or not clean_leading_empty(
        sections_docstr["_title"]
    ):
        sections_docstr["_title"] = make_title_from_func_name(fn_name)
    if "_parameters" in sections_docstr:
        # NOTE: only add/update param info if already any is provided in the doc
        for arg in args:
            arg_name = arg["name"].lstrip("*")
            if (
                arg_name
                not in [
                    param["name"].lstrip("*")
                    for param in sections_docstr["_parameters"]
                ]
                and arg_name != "self"  # Do not document self
            ):
                sections_docstr["_parameters"] = [arg] + sections_docstr["_parameters"]
            else:
                for param in sections_docstr["_parameters"]:
                    if param["name"] == arg["name"].lstrip("*"):
                        # Priority to the docstring
                        for key in arg:
                            if arg[key]:
                                param[key] = arg[key]
    if "_returns" in sections_docstr and len(sections_docstr["_returns"]) in (
        len(rtypes),
        0,
    ):
        # NOTE: only add/update return info if already any is provided in the doc
        if sections_docstr["_returns"]:
            for i, return_arg in enumerate(sections_docstr["_returns"]):
                if "type" not in return_arg or not return_arg["type"]:
                    return_arg["type"] = rtypes[i]
        else:
            sections_docstr["_returns"] = [{"type": rtype} for rtype in rtypes]
    return sections_docstr


def find_range_matching(
    ranges_def: List[List[int]],
    ranges_docstr: List[List[int]],
    lines: Optional[List[str]] = None,
) -> List[List[int]]:
    """Find the ranges in ranges_docstr that matches ranges_def.

    Parameters
    ----------
    ranges_def : List[List[int]]
        Range of signature lines.
    ranges_docstr : List[List[int]]
        Ranges of all the docstrings of the file.
    lines : Optional[List[str]], optional
        Lines of the whole file. Only used to find class docstrings.
        By default, not specified (None).

    Returns
    -------
    corresp_ranges_docstr : List[List[int]]
        Ranges of the docstrings that matches the ranges in ranges_def.
    """
    corresp_ranges_docstr = [[-1, -1] for _ in ranges_def]
    # Check the above line
    for i_def, range_def in enumerate(ranges_def):
        for i_doc, range_docstr in enumerate(ranges_docstr):
            if range_docstr[0] - range_def[1] == 0:
                corresp_ranges_docstr[i_def] = range_docstr
                del ranges_docstr[i_doc]
                continue
    # If not in the above line: check the above line again
    for i_def, range_def in enumerate(ranges_def):
        if corresp_ranges_docstr[i_def] == [-1, -1]:
            for i_doc, range_docstr in enumerate(ranges_docstr):
                if range_docstr[0] - range_def[1] == 1:
                    corresp_ranges_docstr[i_def] = range_docstr
                    del ranges_docstr[i_doc]
                    continue
    # Case where the function is __init__ and the docstring can be
    # right under the class definition
    for i_def, range_def in enumerate(ranges_def):
        if (
            lines
            and corresp_ranges_docstr[i_def] == [-1, -1]
            and "def __init__" in lines[range_def[0]]
        ):
            sorted_ranges = sorted(
                [rang for rang in ranges_docstr if rang[1] <= range_def[0]],
                key=lambda x: x[0],
                reverse=True,
            )
            if sorted_ranges and are_lines_class_head(
                lines[sorted_ranges[0][1] : range_def[0]]
            ):
                corresp_ranges_docstr[i_def] = sorted_ranges[0]
                del ranges_docstr[ranges_docstr.index(sorted_ranges[0])]
    return corresp_ranges_docstr


def are_lines_class_head(lines: List[str]) -> bool:
    """Return if the lines are the head of a class."""
    lines = lines.copy()
    lines = [line.rsplit("#", maxsplit=1)[0].rstrip() for line in lines]
    indent = find_indent(lines)
    return all(not line or line[indent] != " " for line in lines)


def make_title_from_func_name(fn_name: str) -> List[str]:
    """Make title from function name."""
    fn_name = fn_name.strip("_")
    title = " ".join(fn_name.split("_")).capitalize()
    if " " not in title:
        title += " function"
    title += ".\n"
    return [title]



"""Test docstring building."""

import pytest
import pytest_check as check

from docstripy.build_doc.main_builder import build_docstring


@pytest.fixture()
def sections_dict() -> dict:
    """Return sections_dict."""
    sections_dict = {
        "_escaped": True,
        "_title": ["This is a title.\n", "\n", "This is a subtitle.\n"],
        "_parameters": [
            {
                "name": "param1",
                "type": ":obj:`int`",
                "description": [
                    "This is a parameter.\n",
                    "The description can be multiline, and it can be very very long. "
                    "Even longer than this.\n",
                    "\n",
                    "And it can have a default value.\n",
                    "The default value can be very long too. It is an int, and it was "
                    "discovered by a very long process.\n",
                ],
                "optional": True,
                "default": "0",
            },
            {
                "name": "param2",
                "type": "str",
                "description": ["This is another parameter.\n"],
                "optional": True,
                "default": "''",
            },
        ],
        "_raises": [
            {
                "name": "ValueError",
                "description": ["If something goes wrong.\n"],
            }
        ],
        "_returns": [
            {
                "type": "int",
            },
            {
                "name": "result2",
                "type": "str",
                "description": ["The second result.\n"],
            },
        ],
        "_attributes": [
            {
                "name": "attr1",
                "type": "int",
                "description": ["This is an attribute.\n"],
                "default": "0",
            },
        ],
        "Example": [
            "This is an example.\n",
            ">>> function(1, '\\n')\n",
            "0\n",
            "\n",
            "This is another example.\n",
            ">>> function(2, '\\n')\n",
            "0\n",
        ],
        "Notes": ["This is a note.\n"],
    }
    return sections_dict


def test_build_docstring(sections_dict: dict) -> None:
    """Test build docstring for every styles."""
    with open("tests/files/docstr1.txt", "r", encoding="utf-8") as file:
        docstrings = file.readlines()
    # Case include type
    expected_docstrs = {
        "rest": docstrings[:31],
        "google": docstrings[32:68],
        "numpy": docstrings[69:115],
    }
    docstr_config: dict = {"max_len": 76, "indent": 4, "include_type": True}
    for style, expected_docstr in expected_docstrs.items():
        docstr_config["style"] = style
        docstring = build_docstring(
            sections=sections_dict, docstr_config=docstr_config, indent_base=0
        )
        check.equal(
            docstring,
            expected_docstr,
            f"\n**Style {style}**\n**Expected:**\n{''.join(expected_docstr)}"
            f"\n**Got:**\n{''.join(docstring)}",
        )
    # Case NOT include type
    expected_docstrs = {
        "rest": docstrings[116:142],
        "google": docstrings[143:177],
        "numpy": docstrings[178:224],
    }
    docstr_config = {"max_len": 76, "indent": 4, "include_type": False}
    for style, expected_docstr in expected_docstrs.items():
        docstr_config["style"] = style
        docstring = build_docstring(
            sections=sections_dict, docstr_config=docstr_config, indent_base=0
        )
        check.equal(
            docstring,
            expected_docstr,
            f"\n**Style {style}**\n**Expected:**\n{''.join(expected_docstr)}"
            f"\n**Got:**\n{''.join(docstring)}",
        )


def test_build_special() -> None:
    """Test build doc with special cases."""
    sec_dict = {
        "_escaped": False,
        "_title": ["This is a title.\n", "\n", "This is a subtitle.\n"],
        "_parameters": [
            {
                "name": "param1",
                "type": "int",
                "optional": True,
                "default": "0",
            }
        ],
        "_returns": [
            {
                "name": "",
                "type": "str",
            },
        ],
    }
    with open("tests/files/docstr2.txt", encoding="utf-8") as txt_file:
        file_lines = txt_file.readlines()
    docstr_config = {"max_len": 76, "indent": 4, "style": "rest", "include_type": True}
    docstring_rest = build_docstring(
        sections=sec_dict, docstr_config=docstr_config, indent_base=0
    )
    check.equal(docstring_rest, file_lines[0:7])
    docstr_config["style"] = "google"
    docstring_google = build_docstring(
        sections=sec_dict, docstr_config=docstr_config, indent_base=0
    )
    check.equal(docstring_google, file_lines[7:17])
    docstr_config["style"] = "numpy"
    docstring_numpy = build_docstring(
        sections=sec_dict, docstr_config=docstr_config, indent_base=0
    )
    check.equal(docstring_numpy, file_lines[17:30])
    sec_dict = {
        "_escaped": False,
        "_title": ["This is a short title.\n"],
    }
    docstring = build_docstring(
        sections=sec_dict, docstr_config=docstr_config, indent_base=0
    )
    check.equal(docstring, ['"""This is a short title."""\n'])
    sec_dict = {
        "_escaped": False,
        "_title": [
            "This is a very very very long title. Lorem ipsum dolor "
            "sit amet, consectetur adipiscing elit.\n"
        ],
    }
    docstr_config = {"max_len": 50, "indent": 4, "style": "numpy", "include_type": True}
    docstring = build_docstring(
        sections=sec_dict, docstr_config=docstr_config, indent_base=0
    )
    check.equal(
        docstring,
        [
            '"""This is a very very very long title.\n',
            "\n",
            "Lorem ipsum dolor sit amet, consectetur adipiscing\n",
            "elit.\n",
            '"""\n',
        ],
    )



# type: ignore
"""A test module.

Contains some functions with no docstrings.
"""


def clean_trailing_spaces() -> None:  # noqa: D103
    pass


def add_indentation_spaces() -> None:  # noqa: D103
    pass


def remove_end_of_line() -> None:  # noqa: D103
    pass


def add_end_of_line() -> None:  # noqa: D103
    pass



"""Test signature line parsing."""

import pytest_check as check

from docstripy.parse_doc.parse_signature import parse_signature


def test_parse_signature() -> None:
    """Test parse_args function."""
    lines = [
        "def my_func(\n",
        "    arg1: int, # noqa\n",
        "    arg2: str = '5',\n",
        "    *,\n",
        "    arg3=5.0,\n",
        "    **kwargs: Any,\n",
        ") -> Tuple[List[Dict[str, Any]] | None, Tuple[Int, Float]]: # noqa\n",
    ]
    fn_name, rtype, args = parse_signature(lines)
    check.equal(fn_name, "my_func")
    check.equal(rtype, ["List[Dict[str, Any]] | None", "Tuple[Int, Float]"])
    expected_args = [
        {
            "name": "arg1",
            "type": "int",
            "default": "",
            "optional": False,
        },
        {
            "name": "arg2",
            "type": "str",
            "default": "'5'",
            "optional": True,
        },
        {
            "name": "arg3",
            "type": "",
            "default": "5.0",
            "optional": True,
        },
        {
            "name": "**kwargs",
            "type": "Any",
            "default": "",
            "optional": False,
        },
    ]
    check.equal(len(args), len(expected_args))
    for i, arg in enumerate(args):
        check.equal(arg, expected_args[i], f"Error at arg {i}")
    lines = ["def my_func(n):\n"]
    fn_name, rtype, args = parse_signature(lines)
    check.equal(fn_name, "my_func")
    check.equal(rtype, [])
    expected_args = [
        {
            "name": "n",
            "type": "",
            "default": "",
            "optional": False,
        },
    ]
    check.equal(args, expected_args)



"""Parse sections."""
from typing import Dict, List, Tuple

from docstripy.google.parse_doc import (
    parse_sections_ranges as parse_google_sections_ranges,
)
from docstripy.google.parse_doc import (
    parse_wild_sections_ranges as parse_wild_sections_google_ranges,
)
from docstripy.lines_routines import remove_indent
from docstripy.numpy.parse_doc import (
    parse_sections_ranges as parse_numpy_sections_ranges,
)
from docstripy.numpy.parse_doc import (
    parse_wild_sections_ranges as parse_wild_sections_numpy_ranges,
)
from docstripy.rest.parse_doc import parse_sections_ranges as parse_rest_sections_ranges


def parse_sections_ranges(lines: List[str]) -> Tuple[Dict, str]:
    """Parse sections of a docstring and detect docstring style.

    Parameters
    ----------
    lines : List[str]
        Lines of the docstring.

    Returns
    -------
    sec_ranges : Dict
        Section ranges.
        Example: {'_title': [0, 2], '_parameters': [3, 5], "Note": [6, 9]}.
    style : str
        Style of the docstring (one of 'numpy', 'google' or 'rest').
    """
    lines = remove_indent(lines)
    sec_ranges = {"_title": [0, -1]}
    parse_funcs = {
        "google": parse_google_sections_ranges,
        "rest": parse_rest_sections_ranges,
        "numpy": parse_numpy_sections_ranges,
    }
    style = "numpy"
    # NOTE: order google -> rest -> numpy (each section overrides
    # the previous one in case of conflict). The style is
    # determine in this order, overriding previous ones. Default numpy.
    for style_name in ("google", "rest", "numpy"):
        ranges = parse_funcs[style_name](lines)
        ranges = clean_section_ranges(ranges)
        if len(ranges) > 0:
            style = style_name
        sec_ranges.update(ranges)
    if style == "numpy":  # Found numpy section(s) => numpy-doc style
        sec_ranges.update(parse_wild_sections_numpy_ranges(lines))
    else:
        sec_ranges.update(parse_wild_sections_google_ranges(lines))
    sec_ranges = delimit_section_ranges(sec_ranges, len(lines))
    return sec_ranges, style


def clean_section_ranges(ranges: Dict) -> Dict:
    """Clean section ranges.

    If a section is empty, remove it.
    """
    items = list(ranges.items())
    for name, rng in items:
        # Remove empty section
        if rng[0] == -1:
            del ranges[name]
    return ranges


def delimit_section_ranges(ranges: Dict, last_ind: int) -> Dict:
    """Delimit section ranges.

    Determine the end of each section if not already found.
    """
    starts = sorted([rng[0] for rng in ranges.values()])
    for rng in ranges.values():
        if rng[1] == -1:
            ind_sec = starts.index(rng[0])
            end = last_ind if ind_sec == len(starts) - 1 else starts[ind_sec + 1]
            rng[1] = end
    return ranges



"""Parser for ReST style docstrings."""

import re
from typing import Dict, List

from docstripy.lines_routines import find_prefix, remove_indent


def parse_params(
    lines: List[str], section_name: str = "param", pattern_type: str = "type"
) -> List[dict]:
    """Parse parameter section from reStructuredText format.

    Parameters
    ----------
    lines : List[str]
        Lines of docstring.
    section_name : str, optional
        Name of the section, one of "param", "return", "attribute", "raises".
        It is also the pattern before element name for description:
        `:<pattern> name: <description>`. By default, "param".
    pattern_type : str, optional
        Pattern before element name for type: `:<pattern> name: <type>`.
        By default, "type".
    """
    params_list: List[dict] = []
    if section_name == "attribute":
        # For attributes, we can have multiple sections names
        section_names: List[str] = ["ivar", "var", "cvar"]
    else:
        section_names = [section_name]
    for line in lines:
        if line.startswith(tuple(":" + name for name in section_names)):
            sec_name = next(
                name for name in section_names if line.startswith(":" + name)
            )
            len_pattern_1 = len(":" + sec_name)
            len_pattern_2 = len(":" + pattern_type)
            split_dots = re.split(r":\s|\n", line[len_pattern_1:], maxsplit=1)
            param_name = split_dots[0].strip()
            description = split_dots[1].lstrip() if len(split_dots) > 1 else ""
            params_list.append({})
            params_list[-1]["name"] = param_name
            params_list[-1]["type"] = ""  # by default
            params_list[-1]["optional"] = False  # by default
            params_list[-1]["description"] = []
            if description not in ("", "\n"):
                params_list[-1]["description"] = [description]
        elif line.startswith(":" + pattern_type):
            split_dots = re.split(r":\s|\n", line[len_pattern_2:], maxsplit=1)
            param_name = split_dots[0].strip()
            type_p = split_dots[1].strip() if len(split_dots) > 1 else ""
            if type_p.endswith(", optional"):
                type_p = type_p[:-10].strip()
                optional = True
            elif type_p.endswith("optional"):
                type_p = type_p[:-8].strip()
                optional = True
            else:
                type_p = type_p.strip()
                optional = False
            if len(params_list) == 0 or params_list[-1]["type"] != "":
                # New param without description
                params_list.append({})
                params_list[-1]["name"] = param_name
                params_list[-1]["description"] = []
            params_list[-1]["type"] = type_p
            params_list[-1]["optional"] = optional
        elif len(params_list) > 0:
            params_list[-1]["description"].extend(remove_indent([line]))
    return params_list


def parse_sections_ranges(lines: List[str]) -> Dict:
    """Parse reStructuredText sections to the docstring."""
    params_start, params_end = find_prefix(
        lines, (":param", ":type"), (":param", ":type", "\n", " ")
    )
    raises_start, raises_end = find_prefix(lines, (":raises",), (":raises", "\n", " "))
    returns_start, returns_end = find_prefix(
        lines, (":return", ":rtype"), (":return", ":rtype", "\n", " ")
    )
    yields_start, yields_end = find_prefix(
        lines, (":yield",), (":yield", ":rtype", "\n", " ")
    )
    attr_start, attr_end = find_prefix(
        lines,
        (":ivar", ":var", ":cvar"),
        (":ivar", ":var", ":cvar", ":type", "\n", " "),
    )
    return {
        "_parameters": [params_start, params_end],
        "_raises": [raises_start, raises_end],
        "_returns": [returns_start, returns_end],
        "_yields": [yields_start, yields_end],
        "_attributes": [attr_start, attr_end],
    }



"""Preprocessing functions for some sections."""

from typing import List

from docstripy.line_break import line_break
from docstripy.lines_routines import split_first_line


def preprocess_title_build(lines: List[str], max_len: int) -> List[str]:
    """Post-process the title section."""
    lines = split_first_line(lines, max_line_length=max_len)
    lines = line_break(lines, max_line_length=max_len)
    return lines



"""Parse def lines (function signatures)."""

from typing import List

from docstripy.lines_routines import clean_comment


def parse_signature(lines: List[str]) -> tuple[str, List[str], List[dict]]:
    """Parse function signature ("def ...")."""
    lines = clean_comment(lines)
    lines = [line.strip() for line in lines]
    sign_line = " ".join(lines).strip()
    parenthesis1_split = sign_line.split("(", maxsplit=1)
    fn_name = parenthesis1_split[0].replace("def ", "").strip()
    try:
        parenthesis2_split = parenthesis1_split[1].rsplit(")", maxsplit=1)
        if "->" in parenthesis2_split[1]:
            rtype_str = parenthesis2_split[1].replace("->", "").replace(":", "").strip()
            if rtype_str.startswith("Tuple["):
                rtype_str = rtype_str[6:]
                rtype_str = rtype_str.rsplit("]", maxsplit=1)[0].strip()
                rtypes = split_comma(rtype_str)
            else:
                rtypes = [rtype_str]

        else:
            rtypes = []
        comma_split = split_comma(parenthesis2_split[0])
        comma_split = [
            split.strip()
            for split in comma_split
            if split.strip() not in ("*", "/", "")
        ]
        args = [parse_signature_args(split) for split in comma_split]
        return fn_name, rtypes, args
    except (IndexError, ValueError):
        return fn_name, [], []


def parse_signature_args(split: str) -> dict:
    """Parse function arguments on signature line."""
    arg_name, arg_type, arg_default = "", "", ""
    if ":" in split:
        arg_name, arg_default_type = split.split(":", maxsplit=1)
        if "lambda " in arg_name:  # lambda function
            arg_name, arg_default = split.split("=", maxsplit=1)
            arg_type = ""
        else:
            arg_default_type = arg_default_type.strip()
            if "=" in split:
                arg_type, arg_default = arg_default_type.split("=", maxsplit=1)
            else:
                arg_type = arg_default_type
                arg_default = ""
    elif "=" in split:
        arg_name, arg_default = split.split("=", maxsplit=1)
        arg_type = ""
    else:
        arg_name = split
    arg_name = arg_name.strip()
    arg_default = arg_default.strip()
    arg_type = arg_type.strip()
    return {
        "name": arg_name,
        "type": arg_type,
        "default": arg_default,
        "optional": arg_default != "",
    }


def split_comma(line: str) -> List[str]:
    """Parse by comma."""
    line = line.strip()
    count_spec_chars = {
        ("(", ")"): 0,
        ("[", "]"): 0,
        ("{", "}"): 0,
    }
    count_quote = {
        '"': 0,
        "'": 0,
    }
    splits = []
    current_split = ""
    for char in line:
        for spec_char in count_spec_chars:
            if char == spec_char[0]:
                count_spec_chars[spec_char] += 1
            elif char == spec_char[1]:
                count_spec_chars[spec_char] -= 1
        for quote in count_quote:
            if char == quote:
                count_quote[quote] += 1
        if (
            char == ","
            and all(count == 0 for count in count_spec_chars.values())
            and all(count % 2 == 0 for count in count_quote.values())
        ):
            splits.append(current_split.strip())
            current_split = ""
        else:
            current_split += char
    if current_split.strip():
        splits.append(current_split.strip())
    return splits



"""Global docstring parsing functions."""

from typing import List, Tuple

from docstripy.file_parser import parse_ranges
from docstripy.lines_routines import (
    clean_leading_empty,
    clean_trailing_empty,
    remove_indent,
    remove_quotes,
)
from docstripy.parse_doc.parse_params import parse_params_all
from docstripy.parse_doc.postprocessing import postprocess_title_parse
from docstripy.parse_doc.section_ranges import parse_sections_ranges
from docstripy.parse_doc.signature import find_range_matching, merge_docstr_signature


def parse_docstring(
    lines: List[str],
    *,
    add_missing: bool = True,
) -> Tuple[List[List[int]], List[dict], List[bool]]:
    """Docstring parser.

    Parameters
    ----------
    lines : List[str]
        Lines of the file.
    add_missing : bool, optional
        Whether to add missing docstrings, by default True.

    Returns
    -------
    ranges_docstr : List[List[int]]
        Ranges of the docstrings.
    sections_list : List[Dict]
        Parsed sections.
    to_insert : List[bool]
        Whether to insert a new docstring or overwrite the existing one.
    """
    ranges_docstr, ranges_def = parse_ranges(lines)
    out_rng_docstr = []  # output docstring ranges
    sections_list = []
    to_insert = []
    corresp_ranges_docstr = find_range_matching(
        ranges_def=ranges_def,
        ranges_docstr=ranges_docstr,
        lines=lines,
    )
    for rng_def, rng_docstr in zip(ranges_def, corresp_ranges_docstr):
        if rng_docstr == [-1, -1]:
            if add_missing:
                # No docstring: add one
                lines_docstr = ["\n"]
                out_rng_docstr.append([rng_def[1], rng_def[1] + 1])
                to_insert.append(True)
            else:
                continue
        else:
            lines_docstr = lines[rng_docstr[0] : rng_docstr[1]]
            out_rng_docstr.append(rng_docstr)
            to_insert.append(False)
        lines_def = lines[rng_def[0] : rng_def[1]]
        sections = parse_all(lines_docstr)
        sections = merge_docstr_signature(sections, lines_def)
        sections_list.append(sections)
    # Case class docstring
    for range_docstr in ranges_docstr:
        if (
            range_docstr not in corresp_ranges_docstr
            and "class" in lines[range_docstr[0] - 1]
        ):
            lines_docstr = lines[range_docstr[0] : range_docstr[1]]
            out_rng_docstr.append(range_docstr)
            to_insert.append(False)
            sections = parse_all(lines_docstr)
            sections_list.append(sections)
    clean_empty_sections(sections_list)
    return out_rng_docstr, sections_list, to_insert


def parse_all(lines_docstr: List[str]) -> dict:
    r"""Parse the whole docstring.

    Parameters
    ----------
    lines_docstr : List[str]
        Lines of the docstring.

    Returns
    -------
    sections : Dict
        Parsed sections. Can contain the following keys:

        * _title (List[str])
        * _parameters (List[Dict])
        * _returns (List[Dict])
        * _attributes (List[Dict])
        * _raises (List[Dict])
        * any other section name found in the docstring (List[str])

    Examples
    --------
    One can parse a docstring as follows:

    .. code-block:: python

        {
            '_title': ['This is a title\n', '\n', 'This is a description\n'],
            '_parameters': [
                {
                    'name': 'param1',
                    'type': 'int',
                    'optional': False,
                    'description': ['Description of param1\n'],
                },
                {
                    'name': 'param2',
                    'type': 'str',
                    'optional': True,
                    'description': ['Description of param2\n'],
                    'default': '""',
                },
            ],
            '_returns': [
                {
                    'type': 'int',
                    'description': ['Description of the return\n'],
                },
            ],
            'Example': ['Example of a section\n'],
        }
    """
    lines, escaped = remove_quotes(lines_docstr)
    sec_ranges, style = parse_sections_ranges(lines)
    # Preprocess sections
    sections: dict = {}
    sections = {
        sec_name: lines[inds[0] : inds[1]] for (sec_name, inds) in sec_ranges.items()
    }
    sections = {
        sec_name: clean_trailing_empty(lines) for (sec_name, lines) in sections.items()
    }
    sections = {
        sec_name: clean_leading_empty(lines) for (sec_name, lines) in sections.items()
    }
    sections = {
        sec_name: remove_indent(lines) for (sec_name, lines) in sections.items()
    }
    # Manage param, return, attribute, raises using parse_params_all
    section_names_for_parse_params_all = {
        "_parameters": "param",
        "_returns": "return",
        "_yields": "yield",
        "_attributes": "attribute",
        "_raises": "raises",
    }
    for section_unders, section_name in section_names_for_parse_params_all.items():
        if section_unders in sections:
            section = parse_params_all(
                sections[section_unders],
                style=style,
                section_name=section_name,
            )
            section = clean_empty_param_section(section)
            sections[section_unders] = section

    # Manage wild sections : remove header
    for section_name in sections:
        if not section_name.startswith("_"):
            header_len = 1
            if style == "numpy" and sections[section_name][1].strip().startswith("---"):
                header_len = 2
            sections[section_name] = sections[section_name][header_len:]
            if style in ("rest", "google"):
                sections[section_name] = remove_indent(sections[section_name])
    # NOTE : not necessary to modify _title section
    sections["_title"] = postprocess_title_parse(sections["_title"])
    sections["_escaped"] = escaped
    return sections


def clean_empty_param_section(sections: List[dict]) -> List[dict]:
    """Clean empty param dict."""
    new_sections = []
    for param_dict in sections:
        if not all(param_dict[key] in (None, "", False, []) for key in param_dict):
            new_sections.append(param_dict)
    return new_sections


def clean_empty_sections(sections_list: List[dict]) -> None:
    """Clean empty param section."""
    for sections in sections_list:
        items = list(sections.items())
        for key, value in items:
            if not value and key.startswith("_") and key not in ("_title", "_escaped"):
                sections.pop(key)



"""Postprocessing functions for some sections."""

from typing import List

from docstripy.lines_routines import (
    add_eol,
    clean_trailing_empty,
    clean_trailing_spaces,
)


def postprocess_title_parse(lines: List[str]) -> List[str]:
    """Post-process the title section."""
    new_lines = lines.copy()
    if not new_lines:
        return new_lines
    if new_lines[0].strip():
        new_lines[0] = new_lines[0].strip(" ")
        new_lines[0] = new_lines[0][0].capitalize() + new_lines[0][1:]
        if (
            (len(new_lines) >= 3 and not new_lines[1].strip()) or len(new_lines) == 1
        ) and new_lines[0][-2] not in (".", "!", "?"):
            new_lines[-1] = new_lines[-1][:-1] + ".\n"
    return new_lines


def postprocess_description(lines: List[str]) -> List[str]:
    """Post process description."""
    lines = clean_trailing_spaces(lines)
    lines = add_eol(lines)
    lines = clean_trailing_empty(lines)
    if lines and len(lines[-1]) > 1 and lines[-1][-2] not in (".", "?", "!"):
        lines[-1] = lines[-1][:-1] + ".\n"
    # Capitalize first letter
    if len(lines) > 0:
        first_line_desc = lines[0]
        if 97 <= ord(first_line_desc[0]) <= 122:
            cap_letter = first_line_desc[0].capitalize()
            lines[0] = cap_letter + first_line_desc[1:]
    return lines


def postprocess_default_val(default: str) -> str:
    """Post process default value string."""
    if default == "":
        return default
    default.replace("\n", " ")  # remove new lines characters
    return default



"""Test title generation."""

import pytest_check as check

from docstripy.build_doc.preprocessing import preprocess_title_build
from docstripy.parse_doc.postprocessing import postprocess_title_parse


def test_process_title() -> None:
    """Test preprocess & postprocess title."""
    all_lines = [
        ["this is a title\n"],
        ["this is a very long title that should be split\n"],
        ["this is a title. This is a description.\n"],
        ["class.meth is method. This is a description.\n"],
        ["Title. Subtitle. Description.\n", "Another description.\n"],
        ["This is a very very very very long title. Description.\n"],
        ["This is... a title. Description.\n"],
        ["E.g a tile. Description.\n"],
        ["Title.  Description.\n"],  # 2 spaces
    ]
    expected = [
        ["This is a title.\n"],
        ["This is a very long title that\n", "should be split\n"],
        ["This is a title.\n", "\n", "This is a description.\n"],
        ["Class.meth is method.\n", "\n", "This is a description.\n"],
        ["Title.\n", "\n", "Subtitle. Description. Another\n", "description.\n"],
        ["This is a very very very very\n", "long title. Description.\n"],
        ["This is... a title.\n", "\n", "Description.\n"],
        ["E.g a tile.\n", "\n", "Description.\n"],
        ["Title.\n", "\n", "Description.\n"],
    ]
    max_len = 30
    for i, lines in enumerate(all_lines):
        lines = preprocess_title_build(lines, max_len)
        lines = postprocess_title_parse(lines)
        check.equal(lines, expected[i], f"Failed for lines at index {i}")
    long_line1 = [
        (
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit "
            "nullam nec nunc nec nunccurabitur lorem ipsum dolor sit amet ipsum dolor. "
            "Description.\n"
        )
    ]
    lines1 = preprocess_title_build(long_line1, 0)
    lines1 = postprocess_title_parse(lines1)
    check.equal(
        lines1,
        [long_line1[0]],
    )
    long_line2 = [
        (
            "Lorem ipsum dolor sit amet. Consectetur adipiscing elit "
            "nullam nec nunc nec nunccurabitur lorem ipsum dolor sit amet ipsum dolor. "
            "Description.\n"
        )
    ]
    lines2 = preprocess_title_build(long_line2, 0)
    lines2 = postprocess_title_parse(lines2)
    check.equal(
        lines2,
        [long_line2[0][:27] + "\n", "\n", long_line2[0][28:]],
    )



"""Main functions for parsing and building docstrings."""

import argparse
import os.path as osp

from docstripy.write import write_file_ipynb, write_file_py, write_files_recursive


def parse_args() -> dict:
    """Command line parser for docstripy."""
    parser = argparse.ArgumentParser()
    parser.add_argument("in_path", help="File path or root directory path.", type=str)
    parser.add_argument(
        "-s", "--style", help="Style of the docstring", type=str, default="numpy"
    )
    parser.add_argument(
        "-o",
        "--out_path",
        help="Output files dir",
        type=str,
        default="",
    )
    parser.add_argument(
        "-w",
        "--overwrite",
        help="Whether to overwrite the files or not",
        action="store_true",
    )
    parser.add_argument(
        "--indent", help="Base indentation size (2, 4, ...)", type=int, default=4
    )
    parser.add_argument(
        "-l",
        "--length",
        help="Maximum lenght of lines in the docstring. By default, no limit.",
        type=int,
        default=-1,
    )
    parser.add_argument(
        "--noadd",
        help="Prevent adding docstrings when missing.",
        action="store_true",
    )
    parser.add_argument(
        "--notype",
        help="Prevent indicating the types of parameters.",
        action="store_true",
    )
    args = parser.parse_args()
    docstr_config = {
        "style": args.style,
        "max_len": args.length,
        "indent": args.indent,
        "add_missing": not args.noadd,
        "include_type": not args.notype,
    }
    if args.out_path and args.overwrite:
        raise ValueError(
            "Cannot use both `--out_path`/`-o` and `--overwrite`/`-w` options."
        )
    if not args.out_path and not args.overwrite:
        raise ValueError(
            "You must specify an output path with `--out_path`/`-o` "
            "or overwrite file(s) with `--overwrite`/`-w` instead."
        )
    return {
        "in_path": args.in_path,
        "out_path": args.out_path,
        "docstr_config": docstr_config,
        "overwrite": args.overwrite,
    }


def main() -> None:
    """Rewrite file(s) docstrings main function."""
    cli_args = parse_args()
    in_path = cli_args["in_path"]
    if osp.isfile(in_path):
        if osp.splitext(in_path)[1] == ".py":
            write_file_py(**cli_args)
        elif osp.splitext(in_path)[1] == ".ipynb":
            write_file_ipynb(**cli_args)
        else:
            raise ValueError(
                f"File extension not supported: {in_path} "
                "(only .py and .ipynb are supported)"
            )
    write_files_recursive(**cli_args)


if __name__ == "__main__":
    main()



# Copyright (c) 2024 Valentin Goldite. All Rights Reserved.
"""Docstripy - Convert any python docstrings with the docstring format you want.

Support Nympydoc, Google and ReStructuredText as output styles.
The input style should be either Numpy, Google, ReST or even a mix of both.
"""



"""Test docstring section parser."""

import pytest_check as check

from docstripy.parse_doc.main_parser import parse_all


def test_parse_section() -> None:
    """Test section parsing."""
    with open("tests/files/test1.py", "r", encoding="utf-8") as file:
        lines = file.readlines()
    sections1 = parse_all(lines[8:24])
    expected_sections = {
        "_escaped": False,
        "_title": ["Return factorial of n.\n"],
        "_parameters": [
            {
                "name": "n",
                "type": "int",
                "optional": False,
                "description": ["Number to compute factorial of.\n"],
                "default": "",
            },
        ],
        "_returns": [
            {
                "name": "",
                "type": "int",
                "description": ["Factorial of n.\n"],
            },
        ],
        "_raises": [{"name": "ValueError", "description": [], "type": ""}],
    }
    for key, val in expected_sections.items():
        check.equal(sections1[key], val, f"error with key {key}")
    sections2 = parse_all(lines[32:33])
    expected_sections = {
        "_escaped": False,
        "_title": ["Return nth Fibonacci number.\n"],
    }
    check.equal(sections2, expected_sections)

    sections3 = parse_all(lines[54:58])
    expected_sections = {
        "_escaped": True,
        "_title": [r"Return square." + "\n", "\n", r"Use \escaped characters." + "\n"],
    }
    check.equal(sections3, expected_sections)



"""Test raises parsing."""

from typing import List

import pytest_check as check

from docstripy.lines_routines import remove_indent
from docstripy.parse_doc.parse_params import parse_params_all


def test_returns() -> None:
    """Test return parsing."""
    with open("tests/files/test3.py", "r", encoding="utf-8") as file:
        lines = file.readlines()
    rest_doc1 = remove_indent(lines[30:32])
    returns_rest = parse_params_all(rest_doc1, style="rest", section_name="return")
    np_doc1 = remove_indent(lines[69:73])
    returns_np = parse_params_all(np_doc1, style="numpy", section_name="return")
    google_doc1 = remove_indent(lines[103:105])
    returns_google = parse_params_all(
        google_doc1, style="google", section_name="return"
    )
    expected_list = [
        {
            "name": "result",
            "type": "int",
            "description": ["The result of the function.\n"],
        },
    ]
    check.equal(returns_rest, expected_list)
    check.equal(returns_np, expected_list)
    check.equal(returns_google, expected_list)

    doc_empty: List[str] = []
    returns_rest = parse_params_all(doc_empty, style="rest", section_name="return")
    returns_np = parse_params_all(doc_empty, style="numpy", section_name="return")
    returns_google = parse_params_all(doc_empty, style="google", section_name="return")
    check.equal(returns_rest, [])
    check.equal(returns_np, [])
    check.equal(returns_google, [])



"""Global docstring building functions."""

from typing import List

from docstripy.build_doc.preprocessing import preprocess_title_build
from docstripy.google.build_doc import build_doc_google
from docstripy.line_break import line_break
from docstripy.numpy.build_doc import build_doc_numpy
from docstripy.rest.build_doc import build_doc_rest


def build_docstring(
    sections: dict,
    docstr_config: dict,
    indent_base: int,
) -> List[str]:
    """Build a docstring."""
    style = docstr_config["style"]
    max_len = docstr_config["max_len"]
    indent = docstr_config["indent"]
    build_fn = {
        "numpy": build_doc_numpy,
        "google": build_doc_google,
        "rest": build_doc_rest,
    }
    sections["_title"] = preprocess_title_build(
        sections["_title"],
        max_len=max_len - indent_base,
    )
    docstring = sections["_title"].copy() if sections["_title"] else ["\n"]
    esc_char = "r" if sections["_escaped"] else ""
    docstring[0] = esc_char + '"""' + docstring[0]
    if len(sections) == 2 and len(docstring) == 1:  # one-line docstring
        docstring = line_break([docstring[0]], max_len - 6)
        if len(docstring) == 1:  # already one-line docstring after break
            docstring[0] = docstring[0][:-1] + '"""\n'
            return docstring
    docstring = build_fn[style](
        current_docstring=docstring,
        sections_dict=sections,
        max_len=max_len - indent_base,
        indent=indent,
        include_type=docstr_config["include_type"],
    )
    return docstring



"""Test entry points of docstripy."""

import os
import shutil
import sys

import nbformat
import pytest
import pytest_check as check

from docstripy.main import main, parse_args
from docstripy.write import write_file_ipynb, write_file_py, write_files_recursive


def test_main() -> None:
    """Test main entry point."""
    old_argv = sys.argv.copy()
    if os.path.exists("tests/tmp"):
        shutil.rmtree("tests/tmp")
    os.system("docstripy --help")
    sys.argv = [
        "docstripy",
        "tests/files",
        "-o",
        "tests/tmp/google",
        "-s",
        "google",
        "--len",
        "88",
        "--indent",
        "2",
    ]
    main()
    sys.argv[3] = "tests/tmp/rest"
    sys.argv[5] = "rest"
    main()
    sys.argv[3] = "tests/tmp/numpy"
    sys.argv[5] = "numpy"
    main()
    for style in ("google", "numpy", "rest"):
        path = f"tests/tmp/{style}"
        check.is_true(os.path.exists(path), f"Directory not found: {path}")
        for i in range(1, 5):
            path = f"tests/tmp/{style}/test{i}.py"
            check.is_true(os.path.exists(path), f"File not found: {path}")
        path = f"tests/tmp/{style}/empty.py"
        check.is_true(os.path.exists(path), f"File not found: {path}")
        path = f"tests/tmp/{style}/class.py"
        check.is_true(os.path.exists(path), f"File not found: {path}")
        path = f"tests/tmp/{style}/notebook.ipynb"
        check.is_true(os.path.exists(path), f"File not found: {path}")
    # Check no docstring
    with open("tests/tmp/google/empty.py", encoding="utf-8") as file:
        lines = file.readlines()
    check.equal(lines[8], '    """Clean trailing spaces."""\n')
    if os.path.exists("tests/tmp"):
        shutil.rmtree("tests/tmp")
    run = os.system(
        "docstripy tests/files/class.py -o tests/tmp/class.py "
        "-s numpy --len 88 --indent 4"
    )
    check.equal(run, 0)
    check.is_true(os.path.exists("tests/tmp/class.py"))
    # Test class docstring
    with open("tests/tmp/class.py", encoding="utf-8") as file:
        lines = file.readlines()
        print("".join(lines))
    check.equal("    attr : int, optional\n", lines[12])
    check.equal("        The attribute. By default, 0.\n", lines[13])
    check.equal("    attr : int\n", lines[28])
    check.equal('    """\n', lines[30])
    check.equal('        """Forward function."""\n', lines[33])
    # Test notebooks
    run = os.system(
        "docstripy tests/files/notebook.ipynb -o tests/tmp/notebook.ipynb "
        "-s numpy --len 88 --indent 4"
    )
    check.is_true(os.path.exists("tests/tmp/notebook.ipynb"))
    with open("tests/tmp/notebook.ipynb", encoding="utf-8") as file:
        file_dict = nbformat.read(file, as_version=nbformat.NO_CONVERT)
    cell0 = file_dict["cells"][0]["source"].split("\n")
    check.equal("    x : int, optional", cell0[5])
    cell1 = file_dict["cells"][1]["source"].split("\n")
    check.equal("    x : str, optional", cell1[5])
    if os.path.exists("tests/tmp"):
        shutil.rmtree("tests/tmp")
    sys.argv = old_argv


def test_errors(capfd: pytest.CaptureFixture) -> None:
    """Test entry point when there are files errors."""
    old_argv = sys.argv.copy()
    # Case error in parsing
    with pytest.raises(ValueError, match="Error found during docstring parsing."):
        write_file_py(
            "tests/wrong_files/file1.py",
            "tests/tmp/file1.py",
            overwrite=False,
            docstr_config={"style": "numpy", "max_len": 88, "indent": 2},
        )
    # Case wrong extension
    with pytest.raises(ValueError, match="Output file must be a .py file.*"):
        write_file_py(
            "tests/files/empty.py",
            "tests/tmp/empty.ipynb",
            overwrite=False,
            docstr_config={"style": "numpy", "max_len": 88, "indent": 2},
        )
    with pytest.raises(ValueError, match="Output file must be a .ipynb file.*"):
        write_file_ipynb(
            "tests/files/notebook.ipynb",
            "tests/tmp/notebook.py",
            overwrite=False,
            docstr_config={"style": "numpy", "max_len": 88, "indent": 2},
        )
    sys.argv = ["docstripy", "tests/files/docstr1.txt", "-o", "tests/tmp/docstr1.txt"]
    with pytest.raises(ValueError, match="File extension not supported.*"):
        main()
    # Case multiple files
    write_files_recursive(
        "tests/wrong_files",
        "tests/tmp",
        overwrite=False,
        docstr_config={"style": "numpy", "max_len": 88, "indent": 2},
    )
    out, _ = capfd.readouterr()
    check.is_true("Error when parsing file(s):\n" in out)
    # Case args errors
    sys.argv = ["docstripy", "tests/files", "-o", "tests/tmp", "-w"]
    with pytest.raises(ValueError, match="Cannot use both.*"):
        parse_args()
    sys.argv = ["docstripy", "tests/files"]
    with pytest.raises(ValueError, match="You must specify an output path.*"):
        parse_args()
    sys.argv = old_argv



"""File parsing functions."""

from typing import List, Tuple


def parse_ranges(lines: List[str]) -> Tuple[List[List[int]], List[List[int]]]:
    """Parse source code lines ranges.

    Parameters
    ----------
    lines : List[str]
        List of lines to parse.

    Returns
    -------
    ranges_docstr : List[List[int]]
        Ranges of lines containing docstrings.
    ranges_def : List[List[int]]
        Ranges of lines containing function definitions.
    """
    ranges_docstr: List[List[int]] = []
    ranges_def: List[List[int]] = []
    current_docstr_range: List[int] = []
    current_def_range: List[int] = []
    in_docstr, double_quotes, in_def = False, None, False
    states = {
        "ranges_docstr": ranges_docstr,
        "ranges_def": ranges_def,
        "current_docstr_range": current_docstr_range,
        "current_def_range": current_def_range,
        "in_docstr": in_docstr,
        "double_quotes": double_quotes,
        "in_def": in_def,
    }
    for ind_line, line in enumerate(lines):
        states = docstring_parse_range(
            line=line,
            ind_line=ind_line,
            states=states,
        )
        if not in_docstr:
            states = def_parse_range(
                line=line,
                ind_line=ind_line,
                states=states,
            )
    ranges_docstr = states["ranges_docstr"]  # type: ignore
    ranges_def = states["ranges_def"]  # type: ignore
    return ranges_docstr, ranges_def


def docstring_parse_range(
    line: str,
    ind_line: int,
    states: dict,
) -> dict:
    """Parse docstring lines range from source code."""
    ranges = states["ranges_docstr"]
    current_range = states["current_docstr_range"]
    in_docstr = states["in_docstr"]
    double_quotes = states["double_quotes"]
    docstring_starters = ('"""', "'''", 'r"""', "r'''")
    strip_line = line.strip()
    strip_wo_comment = line.rsplit("#", maxsplit=1)[0].strip()
    if strip_line.startswith(docstring_starters) and line.startswith(" "):
        # Case one-line docstring
        if (
            strip_line.count('"""') > 1 or strip_line.count("'''") > 1
        ) and not in_docstr:
            ranges.append([ind_line, ind_line + 1])
        else:
            # Case multi-line docstring
            if not in_docstr:
                current_range = [ind_line]
                in_docstr = True
                double_quotes = strip_line.startswith(('"""', 'r"""'))
            elif (double_quotes and strip_wo_comment.endswith('"""')) or (
                not double_quotes and strip_wo_comment.endswith("'''")
            ):
                current_range.append(ind_line + 1)
                ranges.append(current_range)
                in_docstr = False
                double_quotes = None
    elif in_docstr and (
        (double_quotes and strip_wo_comment.endswith('"""'))
        or (not double_quotes and strip_wo_comment.endswith("'''"))
    ):
        current_range.append(ind_line + 1)
        ranges.append(current_range)
        in_docstr = False
        double_quotes = None
    elif in_docstr and is_def_line(strip_line):
        # Probably not in a docstring
        in_docstr = False
        double_quotes = None
        current_range = []
    states["ranges_docstr"] = ranges
    states["current_docstr_range"] = current_range
    states["in_docstr"] = in_docstr
    states["double_quotes"] = double_quotes
    return states


def is_def_line(strip_line: str) -> bool:
    """Return whether the line is a function or class definition."""
    return (
        strip_line.startswith(("def ", "class "))
        and "(" in strip_line
        and strip_line.split("(", maxsplit=1)[0].count(" ") == 1
    )


def def_parse_range(
    line: str,
    ind_line: int,
    states: dict,
) -> dict:
    """Parse signature lines range from source code."""
    ranges = states["ranges_def"]
    current_range = states["current_def_range"]
    in_def = states["in_def"]
    strip_line = line.rsplit("#", maxsplit=1)[0].strip()
    if not in_def and strip_line.startswith("def "):
        if strip_line.endswith(":"):
            # Case one-line def
            ranges.append([ind_line, ind_line + 1])
        else:
            # Case multi-line def
            current_range = [ind_line]
            in_def = True
    elif in_def and strip_line.endswith(":"):
        current_range.append(ind_line + 1)
        ranges.append(current_range)
        in_def = False
    states["ranges_def"] = ranges
    states["current_def_range"] = current_range
    states["in_def"] = in_def
    return states



"""Main functions for parsing and building docstrings."""

import os
import os.path as osp
from typing import List

import nbformat

from docstripy.build_doc.main_builder import build_docstring
from docstripy.difference import apply_diff
from docstripy.lines_routines import add_eol, add_indent, find_indent
from docstripy.parse_doc.main_parser import parse_docstring


def generate_new_file(file_lines: List[str], docstr_config: dict) -> List[str]:
    """Generate new file with the updated docstrings."""
    try:
        add_missing = docstr_config["add_missing"]
        range_docstrs, sections_list, to_insert = parse_docstring(
            file_lines,
            add_missing=add_missing,
        )
    except (
        IndexError,
        ValueError,
        KeyError,
        ArithmeticError,
        IndentationError,
        NameError,
        ValueError,
        TypeError,
    ) as err:
        raise ValueError("Error found during docstring parsing.") from err
    new_lines = []
    for range_doc, sections in zip(range_docstrs, sections_list):
        try:
            indent_base = find_indent(file_lines[range_doc[0] : range_doc[1]])
            # If a blank line found: look for the next non blank line to get indentation
            i = 1
            while indent_base == -1 and range_doc[1] + i <= len(file_lines):
                indent_base = find_indent(file_lines[range_doc[0] : range_doc[1] + i])
            docstring = build_docstring(
                sections=sections,
                docstr_config=docstr_config,
                indent_base=indent_base,
            )
        except (
            IndexError,
            ValueError,
            KeyError,
            ArithmeticError,
            IndentationError,
            NameError,
            ValueError,
            TypeError,
        ) as err:
            raise ValueError(
                f"Error found at lines {range_doc[0]}-{range_doc[1]} "
                "during docstring building. Please check above error."
            ) from err
        docstring = add_indent(docstring, indent_base)
        new_lines.append("".join(docstring))
    file_new_lines = apply_diff(
        ranges=range_docstrs,
        lines=new_lines,
        old_lines=file_lines,
        to_insert=to_insert,
    )
    return file_new_lines


def write_file_py(
    in_path: str,
    out_path: str,
    *,
    overwrite: bool,
    docstr_config: dict,
) -> None:
    """Write new docstrings on a file."""
    if out_path and not out_path.endswith(".py"):
        raise ValueError(f"Output file must be a .py file (found {out_path}).")
    with open(in_path, encoding="utf-8") as file:
        file_lines = file.readlines()
    file_new_lines = generate_new_file(file_lines, docstr_config)
    out_path = in_path if overwrite else out_path
    if osp.dirname(out_path):
        os.makedirs(osp.dirname(out_path), exist_ok=True)
    with open(out_path, "w", encoding="utf-8") as file:
        file.writelines(file_new_lines)


def write_file_ipynb(
    in_path: str,
    out_path: str,
    *,
    overwrite: bool,
    docstr_config: dict,
) -> None:
    """Write new docstrings on a file."""
    if out_path and not out_path.endswith(".ipynb"):
        raise ValueError(f"Output file must be a .ipynb file (found {out_path}).")
    with open(in_path, encoding="utf-8") as file:
        file_dict = nbformat.read(file, as_version=nbformat.NO_CONVERT)
    for i_cell, cell in enumerate(file_dict["cells"]):
        cell_lines = cell["source"].split("\n")
        cell_lines = add_eol(cell_lines)
        cell_new_lines = generate_new_file(cell_lines, docstr_config)
        file_dict["cells"][i_cell]["source"] = "".join(cell_new_lines)
    out_path = in_path if overwrite else out_path
    if osp.dirname(out_path):
        os.makedirs(osp.dirname(out_path), exist_ok=True)
    with open(out_path, "w", encoding="utf-8") as file:
        nbformat.write(file_dict, file)


def write_files_recursive(
    in_path: str,
    out_path: str,
    *,
    overwrite: bool,
    docstr_config: dict,
) -> None:
    """Write new docstrings on all files in a folder."""
    error_paths = []
    write_func = {
        ".py": write_file_py,
        ".ipynb": write_file_ipynb,
    }
    for dir_path, _, file_names in os.walk(in_path):
        for file_name in file_names:
            if file_name.endswith(tuple(write_func.keys())):
                ext = osp.splitext(file_name)[1]
                file_path = osp.join(dir_path, file_name)
                rel_path = osp.relpath(file_path, in_path)
                file_out_path = osp.join(out_path, rel_path)
                try:
                    write_func[ext](
                        in_path=file_path,
                        out_path=file_out_path,
                        overwrite=overwrite,
                        docstr_config=docstr_config,
                    )
                except (IndexError, ValueError):
                    error_paths.append(file_path)
    if error_paths:
        err_message = "Error when parsing file(s):\n"
        err_message += "\n".join(error_paths)
        err_message += (
            "\nRun `docstripy` on those files individually to get error lines."
        )
        print(err_message)
